import functools
import logging
import os
import uuid

from PyQt4.QtCore import *
from PyQt4 import QtGui

import pypatterns.filter as FilterModule
import pypatterns.relational as RelationalModule

import pomsets.parameter as ParameterModule

import pomsets_app.gui.qt as QtModule


DROP_ABOVE = 'above'
DROP_BELOW = 'below'

def reorderParameters(parameterToMove, newParameterPosition, 
                      controller=None):

    if controller is None:
        raise ValueError('controller cannot be None')

    
    definition = controller.definition()
    contextManager = controller.contextManager()
    builder = contextManager.pomsetBuilder()
    
    # get the current parameter positions
    parameters = controller.getParametersToDisplay()
    newParameters = []

    # insert the parameter in the correct position
    parameterToDrop, dropPosition = newParameterPosition
    for parameter in parameters:
        if parameter.id() == parameterToMove.id():
            continue
        if parameter.id() == parameterToDrop.id():
            if dropPosition == DROP_ABOVE:
                newParameters.append(parameterToMove)
            newParameters.append(parameter)
            if dropPosition == DROP_BELOW:
                newParameters.append(parameterToMove)
            continue
        newParameters.append(parameter)
        pass

    # clear and rebuild the parameter orderings
    definition.parameterOrderingTable().clear()
    for sourceParameter, targetParameter in zip(newParameters[:-1], newParameters[1:]):
        builder.addParameterOrdering(
            definition, sourceParameter.id(), targetParameter.id())
        pass
    
    controller.populateParameters()
    return


def dropEvent(event, function=None, table=None):

    if function is None:
        raise ValueError('function cannot be None')
    if table is None:
        raise ValueError('table cannot be None')

    pos = event.pos()

    item = table.itemAt(pos)    
    if item is None:
        # bad drop, ignore
        return

    parameter = item.data(Qt.UserRole).toPyObject()
    dropIndicatorPosition = table.dropIndicatorPosition()
    dip = "onto"
    if dropIndicatorPosition in [QtGui.QAbstractItemView.OnItem,
                                 QtGui.QAbstractItemView.AboveItem]:
        dip = DROP_ABOVE
    elif dropIndicatorPosition == QtGui.QAbstractItemView.BelowItem:
        dip = DROP_BELOW

    draggedParameter = table.currentItem().data(Qt.UserRole).toPyObject()

    function(draggedParameter, (parameter, dip))
    return


class Controller(QtModule.Controller, QtGui.QDialog):

    def __init__(self, *args, **kwds):
        QtGui.QDialog.__init__(self, *args, **kwds)
        QtModule.Controller.__init__(self)
        return

    def shouldAllowEdit(self, value=None):
        if value is not None:
            self._shouldAllowEdit = value
        if not hasattr(self, '_shouldAllowEdit'):
            self._shouldAllowEdit = False
        return self._shouldAllowEdit

    def shouldReconstructParameterOrdering(self, value=None):
        if value is not None:
            self._shouldReconstructParameterOrdering = value
        if not hasattr(self, '_shouldReconstructParameterOrdering'):
            self._shouldReconstructParameterOrdering = False
        return self._shouldReconstructParameterOrdering

    def definition(self, value=None):
        if value is not None:
            self._definition = value
        if not hasattr(self, '_definition'):
            self._definition = False
        return self._definition

    def populate(self):

        widget = self.widget()
        contextManager = self.contextManager()

        if self.shouldAllowEdit():
            self.setWindowTitle('Edit definition')
        else:
            self.setWindowTitle('View definition')

        comboBoxExecutableType = widget.comboBoxExecutableType 
        executableTypes = contextManager.getExecutableTypes()
        comboBoxExecutableType.addItems(executableTypes)

        # update all gui items with the values from the definition
        pomsetContext = self.pomsetContext()
        definition = self.definition()

        textEditName = widget.textEditName
        textEditName.setText(definition.name() or '')
        textEditDescription =widget.textEditDescription
        textEditDescription.setText(definition.description() or '')
        
        executable = definition.executable()
        executableType = contextManager.getTypeForExecutable(executable)
        index = comboBoxExecutableType.findText(executableType)
        if index >= 0:
            comboBoxExecutableType.setCurrentIndex(index)

        self.updateGuiForEditingExecutableType(executableType)

        self.populateParameters()

        if not self.shouldAllowEdit():
            comboBoxExecutableType.setEditable(False)
            comboBoxExecutableType.setEnabled(False)
            textEditName.setReadOnly(True)
            textEditDescription.setReadOnly(True)
        return


    def updateGuiForEditingExecutableType(self, executableType):

        widget = self.widget()
        contextManager = self.contextManager()
        pomsetContext = self.pomsetContext()
        definition = self.definition()


        executable = definition.executable()
        if executable is None or \
           not contextManager.getTypeForExecutable(executable) == executableType:
            executableClass = contextManager.getClassForExecutableType(executableType)
            executable = executableClass()


        visibilityValues = []
        labelValues = {
            widget.labelExecutablePath:"Executable path"
            }
        if executableType == 'Shell':
            # turn off unnecessary widgets and
            # turn on necessary widgets
            visibilityValues = [True, True,
                                True,
                                False, False,
                                False, False,
                                True, True]

            widget.textEditExecutablePath.setText(os.path.sep.join(executable.path()))
            widget.checkBoxStageable.setChecked(executable.stageable())
            widget.textEditStaticArgs.setText(' '.join(executable.staticArgs()))
            pass
        elif executableType == 'Hadoop Jar':
            # turn off unnecessary widgets
            # turn on necessary widgets
            visibilityValues = [False, False,
                                False,
                                True, True,
                                True, True,
                                False, False]
            widget.textEditJarFile.setText(' '.join(executable.jarFile()))
            widget.textEditJarClass.setText(' '.join(executable.jarClass()))
            widget.textEditStaticArgs.setText(' '.join(executable.staticArgs()))
            pass
        elif executableType == 'Hadoop Streaming':
            # turn off unnecessary widgets
            # turn on necessary widgets
            visibilityValues = [False, False,
                                False,
                                False, False,
                                False, False,
                                False, False]
            pass
        elif executableType == 'Hadoop Pipes':
            # turn off unnecessary widgets
            # turn on necessary widgets

            visibilityValues = [True, True,
                                False,
                                False, False,
                                False, False,
                                False, False]
            widget.textEditExecutablePath.setText(os.path.sep.join(executable.pipesFile()))
            pass
        elif executableType == 'Python Eval':
            visibilityValues = [True, True,
                                False,
                                False, False,
                                False, False,
                                False, False]
            delimiter = '.'
            if not all([len(x) for x in executable.path()]):
                delimiter = ''
            labelValues[widget.labelExecutablePath] = "Module/Function path"
            widget.textEditExecutablePath.setText(delimiter.join(executable.path()))
            pass

        childWidgets = [widget.labelExecutablePath,
                        widget.textEditExecutablePath,
                        widget.checkBoxStageable,
                        widget.labelJarFile,
                        widget.textEditJarFile,
                        widget.labelJarClass,
                        widget.textEditJarClass,
                        widget.labelStaticArgs,
                        widget.textEditStaticArgs]
        for visibilityValue, childWidget in zip(visibilityValues,
                                                childWidgets):
            childWidget.setVisible(visibilityValue)
            pass

        for childWidget, labelText in labelValues.iteritems():
            childWidget.setText(labelText)

        if not self.shouldAllowEdit():
            widget.textEditExecutablePath.setReadOnly(True)
            widget.checkBoxStageable.setEnabled(False)
            widget.textEditJarFile.setReadOnly(True)
            widget.textEditJarClass.setReadOnly(True)
            widget.textEditStaticArgs.setReadOnly(True)

        return


    def getParametersToDisplay(self):
        definition = self.definition()
        parametersTable = definition.parametersTable()
        filter = RelationalModule.ColumnValueFilter(
            'parameter',
            FilterModule.ObjectKeyMatchesFilter(
                filter = FilterModule.EquivalenceFilter(ParameterModule.PORT_TYPE_DATA),
                keyFunction=lambda x: x.portType()
            )
        )
        parameters = RelationalModule.Table.reduceRetrieve(
            parametersTable,
            filter,
            ['parameter'],
            []
        )

        # sort the parameters according to ordering
        parameters = ParameterModule.sortParameters(
            parameters, definition.parameterOrderingTable())
        return parameters


    def populateParameters(self):

        widget = self.widget()
        contextManager = self.contextManager()
        pomsetContext = self.pomsetContext()
        definition = self.definition()

        tableParameters = widget.tableParameters
        while tableParameters.rowCount():
            tableParameters.removeRow(0)

        tableParameters.setEditTriggers(QtGui.QAbstractItemView.NoEditTriggers)
        tableParameters.setSelectionBehavior(QtGui.QAbstractItemView.SelectRows)
        tableParameters.setSelectionMode(QtGui.QAbstractItemView.SingleSelection)

        tableParameters.setDragDropOverwriteMode(False)
        if self.shouldAllowEdit():
            tableParameters.setDragDropMode(QtGui.QAbstractItemView.DragDrop)
        tableParameters.setColumnWidth(0, 200)
        tableParameters.setColumnWidth(1, 120)


        parameters = self.getParametersToDisplay()


        tableParameters.setRowCount(len(parameters))
        for rowIndex, parameter in enumerate(parameters):
            name = parameter.id() or ''
            item = QtGui.QTableWidgetItem('name')
            item.setData(Qt.UserRole, parameter)
            item.setData(Qt.DisplayRole, name)
            tableParameters.setItem(rowIndex, 0, item)

            direction = parameter.portDirection()
            if parameter.getAttribute(ParameterModule.PORT_ATTRIBUTE_ISSIDEEFFECT):
                direction = 'output'
            item = QtGui.QTableWidgetItem('direction')
            item.setData(Qt.UserRole, parameter)
            item.setData(Qt.DisplayRole, direction)
            tableParameters.setItem(rowIndex, 1, item)
            pass


        if self.shouldAllowEdit():
            builder = contextManager.pomsetBuilder()
            reorderParametersFunction = functools.partial(
                reorderParameters,
                controller=self)

            tableParameters.dropEvent = functools.partial(
                dropEvent,
                function=reorderParametersFunction,
                table=tableParameters)

        return


    @pyqtSlot(QtGui.QAbstractButton)
    def on_buttonBox_clicked(self, button):
        widget = self.widget()
        buttonRole = widget.buttonBox.buttonRole(button)
        if buttonRole == QtGui.QDialogButtonBox.ApplyRole:
            if self.shouldAllowEdit():
                self.saveValues()
            pass
        if buttonRole == QtGui.QDialogButtonBox.AcceptRole:
            if self.shouldAllowEdit():
                self.saveValues()
            self.accept()
            pass
        if buttonRole == QtGui.QDialogButtonBox.RejectRole:
            self.reject()
            pass
        return

    @pyqtSlot(QModelIndex)
    def on_tableParameters_doubleClicked(self, index):

        parameter = index.data(Qt.UserRole).toPyObject()
        self.displayParameter(parameter)

        return


    @pyqtSlot(str)
    def on_comboBoxExecutableType_currentIndexChanged(self, value):
        try:
            self.updateGuiForEditingExecutableType(value)
        except TypeError, e:
            logging.debug('error on updating gui >> %s' % e)
        return

    @pyqtSlot()
    def on_buttonAdd_clicked(self):

        if not self.shouldAllowEdit():
            return

        contextManager = self.contextManager()
        pomsetContext = self.pomsetContext()
        definition = self.definition()
        widget = self.widget()

        name = '_'.join(['parameter', uuid.uuid4().hex[:3]])

        builder = contextManager.pomsetBuilder()

        attributes = {
            'direction':ParameterModule.PORT_DIRECTION_INPUT
            }

        builder.addPomsetParameter(definition, name, attributes)

        # TODO:
        # find the last parameter in the list
        # add parameter ordering
        tableParameters = widget.tableParameters
        index = tableParameters.rowCount() - 1
        if index >= 0:
            sourceParameter = \
                tableParameters.item(index, 0).data(Qt.UserRole).toPyObject()

            builder.addParameterOrdering(definition, sourceParameter.id(), name)
            pass

        self.populateParameters()

        return


    @pyqtSlot()
    def on_buttonEdit_clicked(self):
        contextManager = self.contextManager()

        tableParameters = self.widget().tableParameters
        parameters = list(set([x.data(Qt.UserRole).toPyObject() 
                               for x in tableParameters.selectedItems()]))

        if len(parameters) is 0:
            return
        # no need to verify that only one was selected
        # because we've already specified that 
        # as a property of the table

        parameterToEdit = parameters[0]

        self.displayParameter(parameterToEdit)

        return


    def displayParameter(self, parameter):

        contextManager = self.contextManager()


        controller = None
        ui = None

        executableType = str(self.widget().comboBoxExecutableType.currentText())
        if executableType == 'Python Eval':
            # import the modules
            import pomsets_app.gui.qt.edit_python_parameter.widget as WidgetModule
            import pomsets_app.gui.qt.edit_python_parameter.controller as ControllerModule

            # initialize the widget and controller
            controller = ControllerModule.Controller(self)
            ui = WidgetModule.Ui_Dialog()
            pass
        else:

            # import the modules
            import pomsets_app.gui.qt.edit_parameter.widget as WidgetModule
            import pomsets_app.gui.qt.edit_parameter.controller as ControllerModule

            # initialize the widget and controller
            controller = ControllerModule.Controller(self)
            ui = WidgetModule.Ui_Dialog()
            pass
        
        controller.shouldAllowEdit(self.shouldAllowEdit())
        controller.contextManager(contextManager)
        controller.pomsetContext(self.pomsetContext())
        controller.parameter(parameter)
        controller.definition(self.definition())

        ui.setupUi(controller)

        controller.widget(ui)

        # populate the dialog
        controller.populate()

        QObject.connect(contextManager.mainWindow(), 
                        SIGNAL("OnPomsetModified(PyQt_PyObject)"),
                        self.populateParameters)

        controller.show()

        return

    @pyqtSlot()
    def on_buttonRemove_clicked(self):

        if not self.shouldAllowEdit():
            return

        contextManager = self.contextManager()
        builder = contextManager.pomsetBuilder()
        pomsetContext = self.pomsetContext()
        pomset = self.definition()

        tableParameters = self.widget().tableParameters
        parametersToRemove = list(
            set([x.data(Qt.UserRole).toPyObject() 
                 for x in tableParameters.selectedItems()]))

        if len(parametersToRemove):

            # clear the parameter orderings
            pomset.parameterOrderingTable().clear()

            # remove the parameter
            for parameter in parametersToRemove:
                builder.removePomsetParameter(pomset, parameter.id())
                pass

            # now recreate the parameter orderings
            # use the existing ordering in the table
            parameters = self.getDisplayedParameters()

            # because the table still contains references
            # to the parameters that we've removed
            # we need to remove them
            # so that we do not re-add removed parameter orderings
            idsOfParametersToRemove = [x.id() for x in parametersToRemove]
            parameters = [x for x in parameters 
                          if not x.id() in idsOfParametersToRemove]

            self.constructTotalOrder(pomset, parameters)

            self.populateParameters()
            contextManager.mainWindow().emit(
                SIGNAL("OnPomsetModified(PyQt_PyObject)"), pomsetContext)

            pass

        return


    def saveValues(self):

        widget = self.widget()
        contextManager = self.contextManager()
        pomsetContext = self.pomsetContext()
        definition = self.definition()
        
        hasBeenModified = False
        
        name = str(widget.textEditName.text())
        if not name == definition.name():
            definition.name(name)
            hasBeenModified = True

        description = str(widget.textEditDescription.text())
        if not description == definition.description():
            definition.description(description)
            hasBeenModified = True


        executableType = str(widget.comboBoxExecutableType.currentText())
        executable = definition.executable()
        currentExecutableType = contextManager.getTypeForExecutable(executable)
        if not executableType == currentExecutableType:
            executableClass = contextManager.getClassForExecutableType(executableType)
            executable = executableClass()
            definition.executable(executable)
            hasBeenModified = True
            
        if executableType == 'Shell':
            executablePath = str(widget.textEditExecutablePath.text()).split(os.path.sep)
            if not executablePath == executable.path():
                executable.path(executablePath)
                hasBeenModified = True

            isStageable = widget.checkBoxStageable.isChecked()
            if not executable.stageable() == isStageable:
                executable.stageable(isStageable)
                hasBeenModified = True

            staticArgs = str(widget.textEditStaticArgs.text()).split(' ')
            if not executable.staticArgs() == staticArgs:
                executable.staticArgs(staticArgs)
                hasBeenModified = True

        elif executableType == 'Hadoop Jar':

            jarFile = str(widget.textEditJarFile.text()).split(' ')
            if not executable.jarFile() == jarFile:
                executable.jarFile(jarFile)
                hasBeenModified = True

            jarClass = str(widget.textEditJarClass.text()).split(' ')
            if not executable.jarClass() == jarClass:
                executable.jarClass(jarClass)
                hasBeenModified = True
            pass

        elif executableType == 'Hadoop Streaming':
            executable.jarFile([contextManager.getHadoopStreamingJar()])
            executable.jarClass([])
            pass

        elif executableType == 'Hadoop Pipes':

            pipesFile = str(widget.textEditExecutablePath.text()).split(os.path.sep)
            if not executable.pipesFile() == pipesFile:
                executable.pipesFile(pipesFile)
                hasBeenModified = True
            pass
        elif executableType == 'Python Eval':
            executablePath = str(widget.textEditExecutablePath.text())
            if len(executablePath) > 0:
                executablePath = executablePath.rsplit('.', 1)
                if len(executablePath) == 1:
                    executablePath.insert(0, '')
            else:
                executablePath = ['', '']
            if not executablePath == executable.path():
                executable.path(executablePath)
                hasBeenModified = True

        if self.shouldReconstructParameterOrdering():
            hasBeenModified = True
            parameters = self.getDisplayedParameters()
            self.constructTotalOrder(definition, parameters)
            pass

        if hasBeenModified:

            pomsetContext.isModified(True)
            
            contextManager.mainWindow().emit(
                SIGNAL("OnPomsetModified(PyQt_PyObject)"), pomsetContext)
        return

    def getDisplayedParameters(self):
        tableParameters = self.widget().tableParameters
        parameters = [tableParameters.item(x, 0).data(Qt.UserRole).toPyObject()
                      for x in range(tableParameters.rowCount())]
        return parameters


    def constructTotalOrder(self, pomset, parameters):

        contextManager = self.contextManager()
        builder = contextManager.pomsetBuilder()

        for sourceParameter, targetParameter in zip(
            parameters[:-1], parameters[1:]):

            builder.addParameterOrdering(
                pomset, sourceParameter.id(), targetParameter.id())
            pass
        return


    # END class Controller
    pass



import logging
import os

from PyQt4.QtCore import *
from PyQt4 import QtGui

import pypatterns.filter as FilterModule

import pomsets.parameter as ParameterModule

import pomsets_app.utils as AppUtilsModule
import pomsets_app.gui.qt as QtModule

class Controller(QtModule.Controller, QtGui.QDialog):

    ENV_LOCAL = 'local'
    ENV_PICLOUD = 'picloud'

    def __init__(self, *args, **kwds):
        QtGui.QDialog.__init__(self, *args, **kwds)
        QtModule.Controller.__init__(self)
        return

    def getPersistedExecuteEnvironment(self):
        widget = self.widget()
        contextManager = self.contextManager()

        pomsetContext, pomsetReferencePath, isEditable = \
            contextManager.currentDisplayedPomsetInfo()

        execEnvs = pomsetContext.executeEnvironments() or []
        pythonExecEnv = 'pomsets.python.PythonEval'
        for key, envPath in execEnvs:
            if key == 'python eval':
                pythonExecEnv = envPath
        
        if pythonExecEnv == 'pomsets.picloud.PythonEval':
            return Controller.ENV_PICLOUD
        return Controller.ENV_LOCAL


    def persistExecuteEnvironment(self, value):
        widget = self.widget()
        contextManager = self.contextManager()

        pomsetContext, pomsetReferencePath, isEditable = \
            contextManager.currentDisplayedPomsetInfo()

        if value == Controller.ENV_PICLOUD:
            pomsetContext.addExecuteEnvironment(
                'python eval',
                'pomsets.picloud.PythonEval')
            pomsetContext.addCommandBuilder(
                'python eval',
                'pomsets.picloud.CommandBuilder')
        else:
            pomsetContext.addExecuteEnvironment(
                'python eval',
                'pomsets.python.PythonEval')
            pomsetContext.addCommandBuilder(
                'python eval',
                'pomsets.python.CommandBuilder')

        pomsetContext.isModified(True)
        contextManager.mainWindow().emit(
            SIGNAL("OnPomsetModified(PyQt_PyObject)"), pomsetContext)

        return


    def populate(self):

        widget = self.widget()
        contextManager = self.contextManager()

        persistedEnv = self.getPersistedExecuteEnvironment()
        if persistedEnv == Controller.ENV_PICLOUD:
            # set to picloud
            widget.executePicloudButton.setChecked(True)
        else:
            # set to local
            widget.executeLocalButton.setChecked(True)
            pass

        return


    @pyqtSlot(QtGui.QAbstractButton)
    def on_buttonBox_clicked(self, button):
        widget = self.widget()
        buttonRole = widget.buttonBox.buttonRole(button)
        if buttonRole == QtGui.QDialogButtonBox.ApplyRole:
            self.saveValues()
            pass
        if buttonRole == QtGui.QDialogButtonBox.AcceptRole:
            self.saveValues()
            self.accept()
            pass
        if buttonRole == QtGui.QDialogButtonBox.RejectRole:
            self.reject()
            pass
        return


    def saveValues(self):

        contextManager = self.contextManager()
        automaton = contextManager.automaton()
        widget = self.widget()
        
        persistedEnv = self.getPersistedExecuteEnvironment()

        if widget.executePicloudButton.isChecked():
            # do something only if changed
            if persistedEnv != Controller.ENV_PICLOUD:
                self.persistExecuteEnvironment(Controller.ENV_PICLOUD)
                pass
            pass
        else:
            # do something only if changed
            if persistedEnv != Controller.ENV_LOCAL:
                self.persistExecuteEnvironment(Controller.ENV_LOCAL)
                pass
            pass


        return


    # END class Controller
    pass



import os

import pomsets_app.controller.automaton as AutomatonModule


class Frame(object):

    def app(self, value=None):
        if value is not None:
            self._app = value
        return self._app


    def displayPomset(self, pomsetContext, pomsetReferencePath=None):
        
        canvas = self.canvas

        contextManager = self.app().contextManager()

        if pomsetContext is None:
            contextManager.currentDisplayedPomsetInfo((None, None, False))
            canvas.breadCrumbs('No active pomset to display')
            canvas.OnRefresh()
            return

        # after this point
        # it is assumed that pomsetContext is not None
        if pomsetReferencePath is None:
            pomsetReferencePath = [pomsetContext.reference()]
        pomsetReference = pomsetContext.reference()

        contextManager.currentDisplayedPomsetInfo(
            (pomsetContext, pomsetReferencePath, True))

        # display the actual pomset in the canvas
        canvas.displayPomset(pomsetReferencePath)

        self.setBreadCrumbs(pomsetContext, pomsetReferencePath)
        
        self.updateMenuBarActions()
        return


    def selectActivePomset(self, pomset):
        
        # currently a placeholder
        # need to figure out the pomset currently being displayed
        # and select that in the GUI
        if self._updateActivePomsetTreeCtrlSelection:
            print "should update selection"

        
        # set the breadcrumbs for the pomset to be displayed
        contextManager = self.app().contextManager()
        activeContext = contextManager.activePomsetContext()

        return


    def setBreadCrumbs(self, pomsetContext, contextPath=None):
        if contextPath is None:
            contextPath = [pomsetContext.reference()]

        canvas = self.canvas
        canvas.breadCrumbs(' > '.join([x.name() for x in contextPath]))
        
        return

    # END class Frame
    pass





# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'pomsets-gui/resources/qt/python configuration dialog.ui'
#
# Created: Tue Aug 24 12:18:41 2010
#      by: PyQt4 UI code generator 4.7
#
# WARNING! All changes made in this file will be lost!

from PyQt4 import QtCore, QtGui

class Ui_dialog(object):
    def setupUi(self, dialog):
        dialog.setObjectName("dialog")
        dialog.resize(279, 180)
        self.buttonBox = QtGui.QDialogButtonBox(dialog)
        self.buttonBox.setGeometry(QtCore.QRect(20, 140, 241, 32))
        self.buttonBox.setOrientation(QtCore.Qt.Horizontal)
        self.buttonBox.setStandardButtons(QtGui.QDialogButtonBox.Cancel|QtGui.QDialogButtonBox.Ok)
        self.buttonBox.setObjectName("buttonBox")
        self.groupBox = QtGui.QGroupBox(dialog)
        self.groupBox.setGeometry(QtCore.QRect(20, 20, 241, 101))
        self.groupBox.setObjectName("groupBox")
        self.executeLocalButton = QtGui.QRadioButton(self.groupBox)
        self.executeLocalButton.setGeometry(QtCore.QRect(20, 30, 103, 20))
        self.executeLocalButton.setChecked(True)
        self.executeLocalButton.setObjectName("executeLocalButton")
        self.executePicloudButton = QtGui.QRadioButton(self.groupBox)
        self.executePicloudButton.setGeometry(QtCore.QRect(20, 60, 161, 20))
        self.executePicloudButton.setObjectName("executePicloudButton")

        self.retranslateUi(dialog)
        QtCore.QObject.connect(self.buttonBox, QtCore.SIGNAL("accepted()"), dialog.accept)
        QtCore.QObject.connect(self.buttonBox, QtCore.SIGNAL("rejected()"), dialog.reject)
        QtCore.QMetaObject.connectSlotsByName(dialog)

    def retranslateUi(self, dialog):
        dialog.setWindowTitle(QtGui.QApplication.translate("dialog", "Python", None, QtGui.QApplication.UnicodeUTF8))
        self.groupBox.setTitle(QtGui.QApplication.translate("dialog", "Execute Python functions:", None, QtGui.QApplication.UnicodeUTF8))
        self.executeLocalButton.setText(QtGui.QApplication.translate("dialog", "Locally", None, QtGui.QApplication.UnicodeUTF8))
        self.executePicloudButton.setText(QtGui.QApplication.translate("dialog", "Using PiCloud", None, QtGui.QApplication.UnicodeUTF8))


if __name__ == "__main__":
    import sys
    app = QtGui.QApplication(sys.argv)
    dialog = QtGui.QDialog()
    ui = Ui_dialog()
    ui.setupUi(dialog)
    dialog.show()
    sys.exit(app.exec_())




import functools
import logging
import os
import platform
import sys
import uuid

import numpy

from PyQt4 import QtGui
from PyQt4.QtCore import *
from PyQt4.QtOpenGL import *


import pypatterns.relational as RelationalModule
import pypatterns.filter as FilterModule

import pomsets.context as ContextModule
import pomsets.library as LibraryModule
import pomsets.task as TaskModule

import pomsets_app.utils as AppUtilsModule
import pomsets_app.controller.automaton as AutomatonModule
import pomsets_app.gui.frame as FrameModule
import pomsets_app.gui.utils as GuiUtilsModule
import pomsets_app.gui.qt.canvas as CanvasModule
import pomsets_app.gui.qt.menu as MenuModule
import pomsets_app.gui.qt.pomset_tree.widget as PomsetTreeModule


class Frame(QtGui.QMainWindow, FrameModule.Frame):

    def __init__(self):
        QtGui.QMainWindow.__init__(self)
        self.setWindowTitle('pomsets')
        # self.setMinimumSize(1024, 745)
        self.setMinimumSize(800, 600)
        self._widgets = {}
        return

    def putWidget(self, name, widget):
        self._widgets[name] = widget
        return

    def hasWidget(self, name):
        return name in self._widgets

    def getWidget(self, name):
        return self._widgets[name]


    def initializeMenuBar(self):
        menuBar = self.menuBar()

        pomsetMenu = menuBar.addMenu('Pomsets')
        newPomsetAction = pomsetMenu.addAction('New')
        loadPomsetAction = pomsetMenu.addAction('Load')
        savePomsetAction = pomsetMenu.addAction('Save')
        savePomsetAsAction = pomsetMenu.addAction('Save as')
        closePomsetAction = pomsetMenu.addAction('Close')
        self.connect(newPomsetAction,
                     SIGNAL('triggered()'), self.OnCreateNewTopLevelPomset)
        self.connect(loadPomsetAction, 
                     SIGNAL('triggered()'), self.OnLoadTopLevelPomset)
        self.connect(savePomsetAction, 
                     SIGNAL('triggered()'), self.OnSaveCurrentActivePomset)
        self.connect(savePomsetAsAction, 
                     SIGNAL('triggered()'), self.OnSavePomset)
        self.connect(closePomsetAction, 
                     SIGNAL('triggered()'), self.OnCloseActivePomset)
        self.putWidget('menu create new root pomset', newPomsetAction)
        self.putWidget('menu load root pomset', loadPomsetAction)
        self.putWidget('menu save root pomset', savePomsetAction)
        self.putWidget('menu save root pomset as', savePomsetAsAction)
        self.putWidget('menu close root pomset', closePomsetAction)


        # we only add a separator if not a Mac
        # because on a mac, the exit and about menus
        # get added to the bar at the top of the windw
        if not platform.system() in ['Darwin']:
            pomsetMenu.addSeparator()
        exitApplicationAction = pomsetMenu.addAction('Exit')
        self.connect(exitApplicationAction, 
                     SIGNAL('triggered()'), self.OnExitApplication)

        
        # we only do this if not a Mac
        # because on a mac, the exit and about menus
        # get added to the bar at the top of the windw
        if not platform.system() in ['Darwin']:
            helpMenu = menuBar.addMenu('Help')
            aboutApplicationAction = helpMenu.addAction('About')
        else:
            aboutApplicationAction = pomsetMenu.addAction('About')
        self.connect(aboutApplicationAction, 
                     SIGNAL('triggered()'), self.OnAboutApplication)

        executionMenu = menuBar.addMenu('Execute')
        validateExecutionAction = executionMenu.addAction('Validate')
        previewExecutionAction = executionMenu.addAction('Preview')
        executionMenu.addSeparator()
        stopExecutionAction = executionMenu.addAction('Stop')
        resetExecutionAction = executionMenu.addAction('Reset')
        executionMenu.addSeparator()
        localExecutionAction = executionMenu.addAction('Local')
        cloudExecutionAction = executionMenu.addAction('Cloud')
        self.putWidget('menu execute validate', validateExecutionAction)
        self.putWidget('menu execute preview', previewExecutionAction)
        self.putWidget('menu execute stop', stopExecutionAction)
        self.putWidget('menu execute reset', resetExecutionAction)
        self.putWidget('menu execute local', localExecutionAction)
        self.putWidget('menu execute cloud', cloudExecutionAction)


        self.connect(validateExecutionAction, 
                     SIGNAL('triggered()'), self.OnExecuteValidate)
        self.connect(previewExecutionAction, 
                     SIGNAL('triggered()'), self.OnExecutePreview)
        self.connect(stopExecutionAction, 
                     SIGNAL('triggered()'), self.OnExecuteStop)
        self.connect(resetExecutionAction, 
                     SIGNAL('triggered()'), self.OnExecuteReset)
        self.connect(localExecutionAction, 
                     SIGNAL('triggered()'), self.OnExecuteLocal)
        #self.connect(cloudExecutionAction, 
        #             SIGNAL('triggered()'), self.OnExecuteLocal)
        self.connect(cloudExecutionAction, 
                     SIGNAL('triggered()'), self.OnExecuteCloud)


        configurationMenu = menuBar.addMenu('Configuration')
        cloudpoolAction = configurationMenu.addAction('Cloud pool')
        configurationMenu.addSeparator()
        euca2oolsAction = configurationMenu.addAction('Eucalyptus')
        hadoopAction = configurationMenu.addAction('Hadoop')
        pythonAction = configurationMenu.addAction('Python')

        self.connect(cloudpoolAction, 
                     SIGNAL('triggered()'), self.OnCloudPoolConfiguration)
        self.connect(euca2oolsAction, 
                     SIGNAL('triggered()'), self.OnEuca2oolsConfiguration)
        self.connect(hadoopAction, 
                     SIGNAL('triggered()'), self.OnHadoopConfiguration)
        self.connect(pythonAction, 
                     SIGNAL('triggered()'), self.OnPythonConfiguration)

        return


    def updateMenuBarActions(self):

        contextManager = self.app().contextManager()
        if contextManager is None:
            print "need to verify why it is None"
            return

        pomsetContext, pomsetReferencePath, isEditable = \
            contextManager.currentDisplayedPomsetInfo()

        try:
            closePomsetAction = self.getWidget('menu close root pomset')
            closePomsetAction.setEnabled(pomsetContext is not None)
        except KeyError, e:
            print "key error %s" % e
            pass

        shouldEnableExecute = pomsetContext is not None
        shouldEnableExecuteEuca2ools = pomsetContext is not None
        shouldEnableExecuteStop = False

        if pomsetReferencePath is not None and \
                contextManager.hasTaskForPomsetPath(pomsetReferencePath[:1]):
            shouldEnableExecute = False
            shouldEnableExecuteEuca2ools = False


            # TODO:
            # need to verify that the task has not already completed
            # also need to figure out a way to keep this updated
            task = contextManager.getTaskForPomsetPath(pomsetReferencePath[:1])
            shouldEnableExecuteStop = task.isStoppable()
            pass


        if shouldEnableExecute:
            shouldEnableExecuteStop = False

            try:
                threadPoolId = AutomatonModule.ID_EXECUTE_EUCA2OOLS
                threadPool = contextManager.automaton().getThreadPoolUsingId(
                    threadPoolId)
                shouldEnableExecuteEuca2ools = shouldEnableExecute
                if threadPool.isEmpty():
                    # disable the execute eucalyptus menu item
                    shouldEnableExecuteEuca2ools = False
                    pass
            except ValueError, e:
                # this is probably because the threadpool
                # for euca2ools was not created
                shouldEnableExecuteEuca2ools = False
                pass

            # enable the execute menu items
            # disable the kill execute menu item
            pass

        try:
            executeLocalAction = self.getWidget('menu execute local')
            executeCloudAction = self.getWidget('menu execute cloud')
            executeStopAction = self.getWidget('menu execute stop')
            executeResetAction = self.getWidget('menu execute reset')

            executeLocalAction.setEnabled(shouldEnableExecute)
            executeCloudAction.setEnabled(
                shouldEnableExecute and shouldEnableExecuteEuca2ools)
            #executeCloudAction.setEnabled(True)

            executeStopAction.setEnabled(shouldEnableExecuteStop)

            # enable the reset if execution started
            # but is not currently running, i.e. completed, errored, or killed
            executeResetAction.setEnabled(
                not (shouldEnableExecute or shouldEnableExecuteStop))
        except KeyError, e:
            print "key error %s" % e
            pass
        
        return


    def initializeWidgets(self):

        mainSplitterWidget = QtGui.QSplitter(self)

        librarySplitterWidget = QtGui.QSplitter(mainSplitterWidget)
        librarySplitterWidget.setOrientation(Qt.Vertical)

        activeLibraryTreeWidget = PomsetTreeModule.PomsetTreeWidget(librarySplitterWidget)
        activeLibraryTreeWidget.setHeaderLabels(['Active pomsets'])
        activeLibraryTreeWidget.setHeaderHidden(False)
        librarySplitterWidget.addWidget(activeLibraryTreeWidget)
        self.putWidget('active library tree', activeLibraryTreeWidget)

        toolbox = QtGui.QToolBox(librarySplitterWidget)
        toolbox.addItem(PomsetTreeModule.PomsetTreeWidget(toolbox), 'Loaded definitions')
        toolbox.addItem(PomsetTreeModule.PomsetTreeWidget(toolbox), 'Library definitions')
        toolbox.setCurrentIndex(1)
        librarySplitterWidget.addWidget(toolbox)

        canvas = CanvasModule.Canvas(mainSplitterWidget, 800, 600)
        canvas.contextManager(self.app().contextManager())
        canvas.resetDrawables()
        self.putWidget('graph canvas', canvas)
        self.canvas = canvas
        mainSplitterWidget.addWidget(canvas)

        self.setCentralWidget(mainSplitterWidget)

        # signal from popup menu to create node
        QObject.connect(
            toolbox.widget(0),
            SIGNAL("OnCreateNode(PyQt_PyObject,PyQt_PyObject,PyQt_PyObject,QPoint)"),
            self.OnCreateNode)
        QObject.connect(
            toolbox.widget(1), 
            SIGNAL("OnCreateNode(PyQt_PyObject,PyQt_PyObject,PyQt_PyObject,QPoint)"),
            self.OnCreateNode)
        
        # notify the widgets that a node got created
        QObject.connect(self, SIGNAL("OnNodeCreated(PyQt_PyObject,QPoint)"),
                        canvas.OnNodeCreated)
        QObject.connect(self, SIGNAL("OnNodeCreated(PyQt_PyObject,QPoint)"),
                        activeLibraryTreeWidget.OnRefresh)

        # notify the widgets that an edge got created
        QObject.connect(canvas, SIGNAL("OnEdgeCreated(PyQt_PyObject)"),
                        activeLibraryTreeWidget.OnRefresh)

        # notify the widgets that a node got moved
        QObject.connect(canvas, SIGNAL("OnNodesMoved()"),
                        activeLibraryTreeWidget.OnRefresh)

        # notify the widgets that a node got duplicated
        QObject.connect(self, SIGNAL("OnNodesDuplicated(PyQt_PyObject)"),
                        activeLibraryTreeWidget.OnRefresh)
        QObject.connect(self, SIGNAL("OnNodesDuplicated(PyQt_PyObject)"),
                        canvas.OnNodesDuplicated)

        # notify the widgets that a node got deleted
        QObject.connect(self, SIGNAL("OnNodesDeleted()"),
                        activeLibraryTreeWidget.OnRefresh)
        QObject.connect(self, SIGNAL("OnNodesDeleted()"),
                        canvas.OnRefresh)
        
        # notify the widget that a pomset got moved across libraries
        QObject.connect(self, SIGNAL("OnPomsetMoved()"),
                        toolbox.widget(0).OnRefresh)
        QObject.connect(self, SIGNAL("OnPomsetMoved()"),
                        toolbox.widget(1).OnRefresh)


        # notify the widget that a pomset got created
        QObject.connect(self, SIGNAL("OnPomsetCreated()"),
                        toolbox.widget(0).OnRefresh)
        QObject.connect(self, SIGNAL("OnPomsetCreated()"),
                        toolbox.widget(1).OnRefresh)
        QObject.connect(self, SIGNAL("OnPomsetCreated()"),
                        activeLibraryTreeWidget.OnRefresh)

        # notify the widget that a pomset got loaded
        QObject.connect(self, SIGNAL("OnPomsetLoaded(PyQt_PyObject)"),
                        toolbox.widget(0).OnRefresh)
        QObject.connect(self, SIGNAL("OnPomsetLoaded(PyQt_PyObject)"),
                        toolbox.widget(1).OnRefresh)
        QObject.connect(self, SIGNAL("OnPomsetLoaded(PyQt_PyObject)"),
                        activeLibraryTreeWidget.OnRefresh)

        # notify the widget that a pomset got saved
        QObject.connect(self, SIGNAL("OnPomsetSaved(PyQt_PyObject)"),
                        toolbox.widget(0).OnRefresh)
        QObject.connect(self, SIGNAL("OnPomsetSaved(PyQt_PyObject)"),
                        toolbox.widget(1).OnRefresh)
        QObject.connect(self, SIGNAL("OnPomsetSaved(PyQt_PyObject)"),
                        activeLibraryTreeWidget.OnRefresh)

        # notify the widget that a pomset was closed
        QObject.connect(self, SIGNAL("OnPomsetClosed(PyQt_PyObject)"),
                        toolbox.widget(0).OnRefresh)
        QObject.connect(self, SIGNAL("OnPomsetClosed(PyQt_PyObject)"),
                        toolbox.widget(1).OnRefresh)
        QObject.connect(self, SIGNAL("OnPomsetClosed(PyQt_PyObject)"),
                        activeLibraryTreeWidget.OnRefresh)

        # notify the widget that a pomset got modified
        QObject.connect(self, SIGNAL("OnPomsetModified(PyQt_PyObject)"),
                        toolbox.widget(0).OnRefresh)
        QObject.connect(self, SIGNAL("OnPomsetModified(PyQt_PyObject)"),
                        activeLibraryTreeWidget.OnRefresh)
        QObject.connect(self, SIGNAL("OnPomsetModified(PyQt_PyObject)"),
                        canvas.OnRefresh)

        # TODO:
        # notify that a pomset was submitted for execution

        # TODO:
        # notify that a pomset errored in execution

        # TODO:
        # notify a task execution status changed
        QObject.connect(self, SIGNAL("OnTaskExecutionStatusChanged()"),
                        canvas.updateGL)

        return


    def closeEvent(self, event):
        # call our own exit application function
        self.OnExitApplication()
        event.ignore()
        return


    def setWidgetDataSources(self):

        contextManager = self.app().contextManager()

        mainSplitterWidget = self.centralWidget()

        librarySplitterWidget = mainSplitterWidget.widget(0)
        toolboxWidget = librarySplitterWidget.widget(1)
        
        persistentLibraryTreeWidget = toolboxWidget.widget(1)
        
        persistentLibraryTreeWidget.pomsetLibrary(
            contextManager.persistentLibrary())
        persistentLibraryTreeWidget.emptyDataSourceString('None loaded')
        persistentLibraryTreeWidget.contextManager(contextManager)


        notBootstrapFilter = FilterModule.constructNotFilter()
        notBootstrapFilter.addFilter(
            LibraryModule.getBootstrapLoaderPomsetsFilter())
        persistentLibraryTreeWidget.pomsetFilter(notBootstrapFilter)
        persistentLibraryTreeWidget.shouldEnableDragToCanvas(True)
        persistentLibraryTreeWidget.contextualMenuClass(
            MenuModule.LibraryDefinitionContextualMenu)

        transientLibraryTreeWidget = toolboxWidget.widget(0)
        transientLibraryTreeWidget.contextManager(contextManager)
        transientLibraryTreeWidget.pomsetLibrary(
            contextManager.transientLibrary())
        transientLibraryTreeWidget.emptyDataSourceString('None loaded')
        transientLibraryTreeWidget.shouldIndicateModified(True)
        transientLibraryTreeWidget.shouldEnableDragToCanvas(True)
        transientLibraryTreeWidget.contextualMenuClass(
            MenuModule.LoadedDefinitionContextualMenu)

        activeLibraryTreeWidget = librarySplitterWidget.widget(0)
        activeLibraryTreeWidget.contextManager(contextManager)
        activeLibraryTreeWidget.pomsetLibrary(
            contextManager.activeLibrary())
        activeLibraryTreeWidget.emptyDataSourceString('None active')
        activeLibraryTreeWidget.shouldShowInternals(True)
        activeLibraryTreeWidget.shouldExpand(True)
        activeLibraryTreeWidget.shouldIndicateModified(True)
        activeLibraryTreeWidget.contextualMenuClass(
            MenuModule.ActivePomsetContextualMenu)
        activeLibraryTreeWidget.mouseDoubleClickCallback(
            functools.partial(activeLibraryTreeWidget.displayPomset,
                              frame=self))

        return

    def populateWidgets(self):

        mainSplitterWidget = self.centralWidget()

        librarySplitterWidget = mainSplitterWidget.widget(0)
        toolboxWidget = librarySplitterWidget.widget(1)

        widgets = [librarySplitterWidget.widget(0)] + \
            map(toolboxWidget.widget, range(toolboxWidget.count()))
            
        for widget in widgets:
            widget.populate()

        return


    def OnSaveCurrentActivePomset(self):
        contextManager = self.app().contextManager()
        pomsetContext, pomsetReferencePath, isEditable = \
            contextManager.currentDisplayedPomsetInfo()
        fileName = pomsetContext.url()
        self.OnSavePomset(pomsetContext, fileName)
        return


    def OnSavePomset(self, pomsetContext=None, fileName=None):
        contextManager = self.app().contextManager()

        if pomsetContext is None:
            pomsetContext, pomsetReferencePath, isEditable = \
                contextManager.currentDisplayedPomsetInfo()

        if fileName is None:
            import user
            fileName = QtGui.QFileDialog.getSaveFileName(
                self,
                QString("Save pomset"),
                QString(user.home),
                QString("Pomset workflows (*.pomset)"))

        if fileName and len(fileName):
            ContextModule.savePomsetAs(pomsetContext, str(fileName))
            self.emit(SIGNAL("OnPomsetSaved(PyQt_PyObject)"), pomsetContext)
            return True

        return False


    def OnLoadTopLevelPomset(self):

        contextManager = self.app().contextManager()
        pomsetContext = self.OnLoadPomsetIntoLibrary(
            contextManager.activeLibrary())
        if pomsetContext:
            contextManager.activePomsetContext(pomsetContext)
            self.displayPomset(pomsetContext)
            # self.canvas.OnRefresh()
            self.canvas.updateGL()
        return


    def OnLoadPomsetIntoLibrary(self, library, filter=None):
        import user
        fileName = QtGui.QFileDialog.getOpenFileName(
            self,
            QString("Load pomset"), 
            QString(user.home),
            QString("Pomset workflows (*.pomset)"))

        # TODO:
        # remove this hack that currently
        # allows us to prevent loading nest definitions
        if filter is None:
            filter = FilterModule.TRUE_FILTER

        pomsetContext = None
        if fileName and len(fileName):
            pomsetContext = ContextModule.loadPomset(path=fileName)
            if filter.matches(pomsetContext):
                library.addPomsetContext(pomsetContext)
                self.emit(SIGNAL("OnPomsetLoaded(PyQt_PyObject)"), 
                          pomsetContext)
            else:
                # TODO:
                # need to display an error message
                # to the user
                pass
            pass

        return pomsetContext

    def OnRenameTopLevelPomset(self, newName):

        contextManager = self.app().contextManager()

        pomsetContext, pomsetReferencePath, isEditable = \
            contextManager.currentDisplayedPomsetInfo()

        pomset = pomsetContext.reference()
        pomset.name(newName)

        self.displayPomset(pomsetContext)
        self.canvas.OnRefresh()
        self.emit(SIGNAL("OnPomsetModified(PyQt_PyObject)"), pomsetContext)
        return


    def OnCreateNewTopLevelPomset(self):

        contextManager = self.app().contextManager()
        pomsetContext = contextManager.createNewPomset()
        contextManager.addActivePomsetContext(pomsetContext)
        self.displayPomset(pomsetContext)
        self.canvas.OnRefresh()
        self.emit(SIGNAL("OnPomsetCreated()"))
        return


    def OnCreateNewAtomicPomset(self):

        name, status = QtGui.QInputDialog.getText(
            self,
            QString("Create definition"),
            QString("Name:"))

        # create the node only if the user provided a name
        # and did not cancel the request
        if status:
            contextManager = self.app().contextManager()

            name = str(name)
            context = GuiUtilsModule.createAtomicDefinition(
                contextManager, name=name)

            self.emit(SIGNAL("OnPomsetCreated()"))

            # should now display the editing
            contextManager.mainWindow().OnDisplayDefinition(
                context, [context.reference()], allowEdit=True)

        return

    def OnCreateNewNestPomset(self):
        
        name, status = QtGui.QInputDialog.getText(
            self,
            QString("Create nest"),
            QString("Name:"))

        # create the node only if the user provided a name
        # and did not cancel the request
        if status:
            name = str(name)
            #context = GuiUtilsModule.createNestDefinition(
            #    self.app().contextManager(),
            #    name=name)
            # self.emit(SIGNAL("OnPomsetCreated()"))
            print "should create new nest"

        return


    def OnCloseActivePomset(self, pomsetContext=None, 
                            shouldQueryIfModified=True):

        contextManager = self.app().contextManager()

        if pomsetContext is None:
            pomsetContext, pomsetReferencePath, isEditable = \
                contextManager.currentDisplayedPomsetInfo()
        
        pomsetReference = pomsetContext.reference()

        # query the user to save the modification
        if pomsetContext.isModified() and shouldQueryIfModified:
            pomsetName = pomsetReference.name()

            msgBox = QtGui.QMessageBox(self)
            msgBox.setText("pomset \"%s\" has been modified." % pomsetName)
            msgBox.setInformativeText("Discard changes, save, or cancel close?")
            msgBox.setStandardButtons(
                QtGui.QMessageBox.Discard | 
                QtGui.QMessageBox.Save | 
                QtGui.QMessageBox.Cancel)
            msgBox.setDefaultButton(QtGui.QMessageBox.Cancel)
            ret = msgBox.exec_()

            if ret == QtGui.QMessageBox.Cancel:
                return
            if ret == QtGui.QMessageBox.Save:
                pomsetSaved = self.OnSavePomset(pomsetContext=pomsetContext, 
                                                fileName=pomsetContext.url())
                if not pomsetSaved:
                    # the user cancled the file dialog
                    # so we assume that they want to cancel the save
                    return
                pass
            if ret == QtGui.QMessageBox.Discard:
                # discard does nothing, 
                # we continue with the rest of this function
                pass

            pass

        contextManager.closePomset(pomsetContext)

        activeContext = contextManager.activePomsetContext()
        self.displayPomset(activeContext)
        self.canvas.OnRefresh()

        self.emit(SIGNAL("OnPomsetClosed(PyQt_PyObject)"), pomsetContext)
        return

    def OnDuplicateSelection(self, selection):

        import pomsets_app.gui.graph as GraphModule
        nodes = [x for x in selection if isinstance(x, GraphModule.Node)]
        edges = [x for x in selection if isinstance(x, GraphModule.Edge)]

        # first duplicate the nodes
        nodeMap = self.OnDuplicateNodes(nodes)

        # now connect the edges
        self.OnDuplicateEdges(edges, nodeMap)
        return

    def OnDuplicateEdges(self, edges, nodeMap):
        
        contextManager = self.app().contextManager()

        pomsetContext, pomsetReferencePath, isEditable = \
            contextManager.currentDisplayedPomsetInfo()
        pomsetReference = pomsetReferencePath[-1]
        pomset = pomsetReference.definitionToReference()

        for guiEdge in edges:
            sourceNode = nodeMap[guiEdge.inputNode().nodeData]
            sourceParameterId = guiEdge.inputPort().name
            targetNode = nodeMap[guiEdge.outputNode().nodeData]
            targetParameterId = guiEdge.outputPort().name
            contextManager.pomsetBuilder().connect(
                pomset,
                sourceNode, sourceParameterId,
                targetNode, targetParameterId)
            
        self.emit(
            SIGNAL("OnPomsetModified(PyQt_PyObject)"), pomsetContext)
        return



    def OnDuplicateNodes(self, nodes):

        contextManager = self.app().contextManager()

        pomsetContext, pomsetReferencePath, isEditable = \
            contextManager.currentDisplayedPomsetInfo()
        pomsetReference = pomsetReferencePath[-1]
        pomset = pomsetReference.definitionToReference()

        nodeMap = {}
        for guiNode in nodes:
            dataNode = guiNode.nodeData
            copyDataNode = contextManager.pomsetBuilder().copyNode(
                pomset, dataNode)
            copyDataNode.name('copy of ' + dataNode.name())

            nodeMap[dataNode] = copyDataNode
            pass

        self.emit(SIGNAL("OnNodesDuplicated(PyQt_PyObject)"), nodeMap)
        return nodeMap


    def OnDeleteSelection(self, selection):
        import pomsets_app.gui.graph as GraphModule
        nodes = [x for x in selection if isinstance(x, GraphModule.Node)]
        edges = [x for x in selection if isinstance(x, GraphModule.Edge)]

        # first remove the edges
        self.OnDeleteEdges(edges)

        # then remove the nodes
        self.OnDeleteNodes(nodes)
        return

    def OnDeleteEdges(self, edges):
        contextManager = self.app().contextManager()

        pomsetContext, pomsetReferencePath, isEditable = \
            contextManager.currentDisplayedPomsetInfo()
        pomsetReference = pomsetReferencePath[-1]
        pomset = pomsetReference.definitionToReference()

        for guiEdge in edges:
            sourceNode = guiEdge.inputNode().nodeData
            sourceParameterId = guiEdge.inputPort().name
            targetNode = guiEdge.outputNode().nodeData
            targetParameterId = guiEdge.outputPort().name
            contextManager.pomsetBuilder().disconnect(
                pomset, 
                sourceNode, sourceParameterId,
                targetNode, targetParameterId
                )
            pass

        self.emit(
            SIGNAL("OnPomsetModified(PyQt_PyObject)"), pomsetContext)
        return


    def OnDeleteNodes(self, nodes):
        contextManager = self.app().contextManager()

        pomsetContext, pomsetReferencePath, isEditable = \
            contextManager.currentDisplayedPomsetInfo()
        pomsetReference = pomsetReferencePath[-1]
        pomset = pomsetReference.definitionToReference()

        for guiNode in nodes:
            dataNode = guiNode.nodeData
            contextManager.pomsetBuilder().removeNode(
                pomset, dataNode)
        self.emit(SIGNAL("OnNodesDeleted()"))
        return


    def OnCreateNode(self, pomsetContext, pomsetReference, 
                     definitionToReference, pos, name=None):

        contextManager = self.app().contextManager()

        if name is None:
            # create a node
            # set its position
            name = '%s_%s' % (definitionToReference.name() or "anonymous",
                              uuid.uuid4().hex[:3])

        node = contextManager.pomsetBuilder().createNewNode(
            pomsetReference.definitionToReference(), 
            name=name,
            definitionToReference=definitionToReference)

        pomsetContext.isModified(True)

        # emit a signal
        # so that the other widgets
        # will receive it and update accordingly
        self.emit(SIGNAL("OnNodeCreated(PyQt_PyObject,QPoint)"), node, pos)

        return


    def OnMoveDefinitionToLibrary(self, pomsetContext):
        contextManager = self.app().contextManager()
        persistentLibrary = contextManager.persistentLibrary()
        transientLibrary = contextManager.transientLibrary()
        definition = pomsetContext.pomset()

        # backup the existing library
        AppUtilsModule.backupLibrary(
            persistentLibrary.bootstrapLoaderDefinitionsDir())
        
        definitionFilter = RelationalModule.ColumnValueFilter(
            'definition',
            FilterModule.IdentityFilter(definition)
        )
        
        # remove from transient library
        transientLibrary.removeDefinition(definitionFilter)

        # define the url for the definition
        definitionPath = definition.id() + '.pomset'
        definition.url(definitionPath)
        outputDir = persistentLibrary.bootstrapLoaderDefinitionsDir()
        
        # save out the definition
        ContextModule.pickleDefinition(
            os.path.join(outputDir, definitionPath), definition)
        
        # add to persistent library
        definition.isLibraryDefinition(True)
        persistentLibrary.addPomsetContext(pomsetContext)
        
        # save out the bootstrap loader pomset
        # that will load this definition on the next startup
        persistentLibrary.saveBootstrapLoaderPomset()
        
        # update GUI
        self.emit(SIGNAL("OnPomsetMoved()"))

        return


    def OnDisplayDefinition(self, pomsetContext,
                            pomsetReferencePath, allowEdit=False):

        reference = pomsetReferencePath[-1]
        definitionToReference = reference.definitionToReference()

        if definitionToReference.isAtomic():
            return self.OnDisplayAtomicDefinition(
                pomsetContext, pomsetReferencePath,
                allowEdit=allowEdit)

        return self.OnDisplayNestDefinition(
            pomsetContext, pomsetReferencePath, allowEdit=allowEdit)


    def OnDisplayNestDefinition(self, pomsetContext,
                                pomsetReferencePath, allowEdit=False):

        # TODO:
        # need to be able to specify whether a nest pomset is editable
        #pomsetContext, pomsetReferencePath, isEditable = \
        #    contextManager.currentDisplayedPomsetInfo()

        self.displayPomset(pomsetContext, pomsetReferencePath=pomsetReferencePath)
        return


    def OnDisplayAtomicDefinition(self, pomsetContext, 
                                  pomsetReferencePath, allowEdit=False):

        reference = pomsetReferencePath[-1]

        contextManager = self.app().contextManager()

        # import the modules
        import pomsets_app.gui.qt.edit_definition.widget as WidgetModule
        import pomsets_app.gui.qt.edit_definition.controller as ControllerModule

        # initialize the widget
        controller = ControllerModule.Controller(self)
        controller.shouldAllowEdit(allowEdit)
        controller.contextManager(contextManager)
        controller.pomsetContext(pomsetContext)
        controller.definition(reference.definitionToReference())

        ui = WidgetModule.Ui_Dialog()
        ui.setupUi(controller)

        controller.widget(ui)

        # populate the dialog
        controller.populate()

        controller.show()

        return


    def OnEditParameterBinding(self, pomsetContext, definition, parameter):
        editTaskController = self.OnEditTask(pomsetContext, definition)
        controller = editTaskController.editParameterBinding(parameter)
        return controller


    def OnEditTask(self, pomsetContext, definition):
        """
        The task is actually the ReferenceDefinition
        from which the task will be built
        """

        contextManager = self.app().contextManager()

        # import the modules
        import pomsets_app.gui.qt.edit_task.widget as WidgetModule
        import pomsets_app.gui.qt.edit_task.controller as ControllerModule

        # initialize the widget
        controller = ControllerModule.Controller(self)
        controller.contextManager(contextManager)
        controller.pomsetContext(pomsetContext)
        controller.definition(definition)

        ui = WidgetModule.Ui_Dialog()
        ui.setupUi(controller)

        controller.widget(ui)

        # populate the dialog
        controller.populate()

        controller.show()

        return controller


    def OnDefinitionRemoved(self, library, pomsetContext):
        self.emit(SIGNAL("OnPomsetClosed(PyQt_PyObject)"), 
                  pomsetContext)
        return

    def OnCloudPoolConfiguration(self):

        contextManager = self.app().contextManager()

        # import the modules
        import pomsets_app.gui.qt.cloud_pool.widget as WidgetModule
        import pomsets_app.gui.qt.cloud_pool.controller as ControllerModule

        # initialize the widget
        controller = ControllerModule.Controller(self)
        controller.contextManager(contextManager)

        ui = WidgetModule.Ui_Dialog()
        ui.setupUi(controller)

        controller.widget(ui)

        # populate the dialog
        controller.populate()

        controller.show()

        return


    def OnEuca2oolsConfiguration(self):
        contextManager = self.app().contextManager()

        # import the modules
        import pomsets_app.gui.qt.euca2ools_configuration.widget as WidgetModule
        import pomsets_app.gui.qt.euca2ools_configuration.controller as ControllerModule

        # initialize the widget
        controller = ControllerModule.Controller(self)
        controller.contextManager(contextManager)

        ui = WidgetModule.Ui_Dialog()
        ui.setupUi(controller)

        controller.widget(ui)

        # populate the dialog
        controller.populate()

        controller.show()

        return


    def OnHadoopConfiguration(self):
        contextManager = self.app().contextManager()

        # import the modules
        import pomsets_app.gui.qt.hadoop_configuration.widget as WidgetModule
        import pomsets_app.gui.qt.hadoop_configuration.controller as ControllerModule

        # initialize the widget
        controller = ControllerModule.Controller(self)
        controller.contextManager(contextManager)

        ui = WidgetModule.Ui_Dialog()
        ui.setupUi(controller)

        controller.widget(ui)

        # populate the dialog
        controller.populate()

        controller.show()

        return


    def OnPythonConfiguration(self):
        # need to display a gui for users to determine 
        # if they want to execute Python functions
        # - locally
        # - on picloud
        # - on their own cloud (in the future)
        contextManager = self.app().contextManager()

        # import the modules
        import pomsets_app.gui.qt.python_configuration.widget as WidgetModule
        import pomsets_app.gui.qt.python_configuration.controller as ControllerModule

        # initialize the widget
        controller = ControllerModule.Controller(self)
        controller.contextManager(contextManager)

        # NOTE:
        # for some reason,
        # the dialog class generated on my macbook
        # uses a lower case "d" for Ui_dialog
        ui = WidgetModule.Ui_dialog()
        ui.setupUi(controller)

        controller.widget(ui)

        # populate the dialog
        controller.populate()

        controller.show()


    def OnExecuteValidate(self):
        contextManager = self.app().contextManager()
        pomsetContext, pomsetReferencePath, isEditable = \
            contextManager.currentDisplayedPomsetInfo()

        if pomsetReferencePath is None:
            return

        errorInfo = [x for x in pomsetContext.reference().parameterBindingErrors()]

        msgBox = QtGui.QMessageBox(self)
        msgBox.setTextFormat(Qt.RichText)
        text = "Found %s error" % len(errorInfo)
        if len(errorInfo) is not 1:
            text = text + "s"
        msgBox.setText(text)

        if len(errorInfo) is not 0:
            errors = ["%s" % error for parameter,error in errorInfo]
            informativeText = """<ul><li>%s
</li></ul>""" % "</li><li>".join(errors)
            msgBox.setInformativeText(informativeText)

        msgBox.exec_()
            

        return

    def OnExecutePreview(self):
        contextManager = self.app().contextManager()

        # import the modules
        import pomsets_app.gui.qt.task_preview.widget as WidgetModule
        import pomsets_app.gui.qt.task_preview.controller as ControllerModule

        # initialize the widget
        controller = ControllerModule.Controller(self)
        controller.contextManager(contextManager)

        ui = WidgetModule.Ui_Dialog()
        ui.setupUi(controller)

        controller.widget(ui)

        # populate the dialog
        controller.populate()

        controller.show()

        return


    def OnExecuteReset(self):

        contextManager = self.app().contextManager()
        pomsetContext, pomsetReferencePath, isEditable = \
            contextManager.currentDisplayedPomsetInfo()

        if pomsetReferencePath is not None and \
                contextManager.hasTaskForPomsetPath(pomsetReferencePath[:1]):
            contextManager.removeTaskForPomset(pomsetReferencePath[0])
            pass

        self.displayPomset(pomsetContext, pomsetReferencePath)

        return


    def OnExecuteStop(self):

        contextManager = self.app().contextManager()
        pomsetContext, pomsetReferencePath, isEditable = \
            contextManager.currentDisplayedPomsetInfo()

        if pomsetReferencePath is not None and \
                contextManager.hasTaskForPomsetPath(pomsetReferencePath[:1]):
            task = contextManager.getTaskForPomsetPath(pomsetReferencePath[:1])

            # TODO:
            # verify that the task is not already stopped
            if task.isStoppable():
                task.stop()
            else:
                logging.debug("task is no longer stoppable")
            pass

        return


    def OnExecuteLocal(self):
        self.OnExecutePomset(AutomatonModule.ID_EXECUTE_LOCAL)
        return
    

    def OnExecuteCloud(self):
        self.OnExecutePomset(AutomatonModule.ID_EXECUTE_EUCA2OOLS)
        return


    def OnAboutApplication(self):
        # TODO:
        # should load from file?

        text = """
<body>
<strong>%s</strong> %s<br/>""" % (str(self.app().applicationVersion()),
                                  str(self.app().applicationName())) + \
        u"\u00A9 2010 nephosity llc, michael j pan"+ \
        """<br/>
</body>
"""
        informativeText = """
<body>
<a href="http://pomsets.org">pomsets</a> is workflow management <br/>
for your cloud<br/>
<br/>
<a href="http://pomsets.org/License">License</a><br/>
</body>
"""
        msgBox = QtGui.QMessageBox(self)
        msgBox.setTextFormat(Qt.RichText)
        msgBox.setText(text)
        msgBox.setInformativeText(informativeText)
        msgBox.exec_()
        return
        

    def OnExitApplication(self):

        contextManager = self.app().contextManager()

        # TODO:
        # verify that there are no executing pomsets
        # if there are, then need to query user
        # for cancel exit, or for killing the jobs


        modifiedContexts = []

        mainSplitterWidget = self.centralWidget()
        librarySplitterWidget = mainSplitterWidget.widget(0)
        toolboxWidget = librarySplitterWidget.widget(1)
        transientLibraryWidget = toolboxWidget.widget(0)
        activeLibraryWidget = librarySplitterWidget.widget(0)

        for widget in [activeLibraryWidget, transientLibraryWidget]:
            definitionTable = widget.pomsetLibrary().definitionTable()
            filter = FilterModule.constructAndFilter()
            filter.addFilter(widget.pomsetFilter())
            filter.addFilter(RelationalModule.ColumnValueFilter(
                    'context',
                    FilterModule.ObjectKeyMatchesFilter(
                        FilterModule.IdentityFilter(True),
                        keyFunction = lambda x: x.isModified())))
            modifiedContexts.extend(
                RelationalModule.Table.reduceRetrieve(
                    definitionTable,
                    filter, ['context'], []))
            pass

        canExit = len(modifiedContexts) is 0

        if not canExit:
            msgBox = QtGui.QMessageBox(self)
            msgBox.setText("There are unsaved, modified pomsets.")
            msgBox.setInformativeText("Discard changes or cancel exit?")
            msgBox.setStandardButtons(
                QtGui.QMessageBox.Discard | 
                QtGui.QMessageBox.Cancel)
            msgBox.setDefaultButton(QtGui.QMessageBox.Cancel)
            ret = msgBox.exec_()
            if ret == QtGui.QMessageBox.Discard:
                canExit = True
            
        if canExit:
            sys.exit(0)

        return
    

    def OnExecutePomset(self, executionEnvId):

        contextManager = self.app().contextManager()

        # TODO:
        # need to check and verify
        # that the specified threadpool
        # has at least one allocated thread
        # if not, should query user

        # TODO:
        # check if the pomset has been modified
        # and if so, query the user for save

        try:
            activePomsetContext = contextManager.activePomsetContext()

            commandBuilderMap = contextManager.getDefaultCommandBuilderMap()
            activePomsetContext.overlayCommandBuilders(commandBuilderMap)
            logging.debug("using command builders >> %s" % commandBuilderMap)

            executeEnvironmentMap = contextManager.getDefaultExecuteEnvironmentMap()
            activePomsetContext.overlayExecuteEnvironments(executeEnvironmentMap)
            logging.debug("using execute environments >> %s" % executeEnvironmentMap)

            requestKwds = contextManager.generateRequestKwds(
                executeEnvironmentId = executionEnvId,
                executeEnvironmentMap = executeEnvironmentMap,
                commandBuilderMap = commandBuilderMap
                )

            contextManager.executePomset(
                pomset=activePomsetContext.reference(),
                requestKwds=requestKwds,
                shouldWait=False)

        except Exception, e:
            import traceback
            exceptionType, exceptionValue, exceptionTraceback = sys.exc_info()
            stackTrace = \
                traceback.format_exception(exceptionType, exceptionValue,
                                           exceptionTraceback)
            logging.error("Error executing pomset >> %s" % stackTrace)

            # TODO:
            # show an error box to the user
            messageBox = QtGui.QMessageBox(self)
            # logging.error("Error executing pomset >> %s" % e)
            # messageBox.showMessage("Error executing pomset >> %s" % e)
            messageBox.setText('Execution error')
            messageBox.setInformativeText(str(e))
            messageBox.setIcon(QtGui.QMessageBox.Critical)
            messageBox.setStandardButtons(
                QtGui.QMessageBox.Ok)
            ret = messageBox.exec_()

            pass
        finally:
            # TODO:
            # update the execute menu items
            # i.e. disable execution for this pomset
            # now that execution has happened
            # and also enable kill
            self.updateMenuBarActions()
            pass
            

        return


    # END class Frame
    pass



from PyQt4.QtCore import *
from PyQt4 import QtGui

class Controller(object):

    def contextManager(self, value=None):
        if value is not None:
            self._contextManager = value
        if not hasattr(self, '_contextManager'):
            self._contextManager = None
        return self._contextManager

    def pomsetContext(self, value=None):
        if value is not None:
            self._pomsetContext = value
        if not hasattr(self, '_pomsetContext'):
            self._pomsetContext = None
        return self._pomsetContext

    def pomsetReference(self, value=None):
        if value is not None:
            self._pomsetReference = value
        if not hasattr(self, '_pomsetReference'):
            self._pomsetReference = None
        return self._pomsetReference

    def widget(self, value=None):
        if value is not None:
            self._widget = value
        if not hasattr(self, '_widget'):
            self._widget = None
        return self._widget

    def connectSignals(self):
        raise NotImplementedError('%s.connectSignals' % self.__class__)

    def notifyUserOfError(self, error,
                          text='Error',
                          icon=None):

        # show an error box to the user
        messageBox = QtGui.QMessageBox(self)
        messageBox.setText(text)
        messageBox.setInformativeText(str(error))

        if icon is None:
            icon = QtGui.QMessageBox.Critical
        messageBox.setIcon(icon)

        messageBox.setStandardButtons(
            QtGui.QMessageBox.Ok)
        ret = messageBox.exec_()
        return ret


    # END class Controller
    pass
